/*

Launcher stub for tlshell, a tlmgr gui, with hidden console window
and with UAC prompt, written in 2018 by Siep Kroonenberg.

Based on tlmgr-gui.c for the original Perl/Tk gui in 2014/2016 by
the same author.

We assume Windows Vista or later since TL won't even install on
earlier versions.

Public Domain
*/

#undef _WIN32_WINNT
#define _WIN32_WINNT 0x0600

#include <windows.h>
#include <winnt.h>
#include <shellapi.h>
#include <shlobj.h>
#include <wtypes.h>
#include <wchar.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

static wchar_t * instroot;
static wchar_t * tclkit;
static wchar_t * tlshell;
static char release[5];

#define FILE_EXISTS(f) ( GetFileAttributes(f) != INVALID_FILE_ATTRIBUTES )

#define DIEFLAGS MB_ICONERROR | MB_TASKMODAL | MB_SETFOREGROUND
void die( wchar_t * s, int i ) {
  MessageBox( NULL, s, L"ERROR!", DIEFLAGS );
  PostQuitMessage( i );
  // without some explicit exit, I got a popup ... has stopped working
  ExitProcess( i );
}

// free and set to NULL
#define FREE0(p) { if( p ) { free( p ); p = NULL; }}

void * calloc_fatal( size_t nitems, size_t item_size ) {
  void * p;
  p = calloc( nitems, item_size );
  if ( !p ) ExitProcess( 1 );
  return p;
}

// wide-char [v]asprintf with automatic sizing and allocation
// adapted from source code of asprintf and vasprintf
int tl_vwasprintf( wchar_t ** ret, const wchar_t * fmt, va_list ap ) {
  FILE * fp;
  wchar_t * buf;
  int len;

  fp = fopen( "NUL", "w" );
  if ( ! fp ) {
    *ret = NULL;
    return -1;
  }
  len = vfwprintf( fp, fmt, ap );
  fclose( fp );
  if( len < 0 ) return -1;
  buf = calloc_fatal( len + 1, sizeof( wchar_t ));
  if ( vswprintf( buf, len, fmt, ap ) != len ) {
    FREE0( buf );
    *ret = NULL;
    return -1;
  }
  *ret = buf;
  return len;
}

void tl_wasprintf( wchar_t ** ret, const wchar_t * fmt, ... ) {
  int len;
  va_list ap;
  va_start( ap, fmt );
  len = tl_vwasprintf( ret, fmt, ap );
  va_end( ap );
  if ( len < 0 ) die( L"tl_wasprintf: Failure to construct string", 1 );
}

void get_tl_release() {
  // read release from TL database
  wchar_t * dbfilename;
  char * cptr;
  int i, j, l;
  FILE * dbfile;
  char dbline[200];

  // calculate filename of TL database
  tl_wasprintf( &dbfilename, L"%ls\\tlpkg\\texlive.tlpdb", instroot );

  // read start of database
  dbfile = _wfopen( dbfilename, L"r" );
  if ( !dbfile ) {
    release[0] = '\0';
    return;
  }
  free( dbfilename );
  for ( i=0; i<200; i++ ) {
    memset( dbline, 0, sizeof( dbline ));
    cptr = fgets( dbline, sizeof( dbline ), dbfile );
    if( !cptr ) {
      // i.e. EOF or error
      fclose( dbfile );
      release[0] = '\0';
      return;
    }
    l = strlen( "depend release/" );
    if( !strncmp( "depend release/", dbline, l)) {
      // next 4 chars should be digits
      for( j=0; j<4; j++ ) {
        release[j] = dbline[l+j];
        if( release[j] < '0' || release[j] > '9' ) {
          memset( release, 0, sizeof( release ));
          break;
        }
      }
      release[4] = '\0';
      fclose( dbfile );
      return;
    }
  }
  // if we got here, release should still consist
  // of zeroes from the initial default initialization
  fclose( dbfile );
  return;
}

BOOL in_HKLM() {
  /*
  Check for uninstall reg entry in HKLM for current TL.

  On w64, registry entries under HKLM\software\wow6432node will be
  transparently aliased to corresponding entries under HKLM\software
  so for 32-bits programs so no worries on this score.

  For just checking the existence of the uninstall registry key,
  we can get away with ascii/ansi/utf8.

  */
  long result;
  HKEY uninst_key;
  char keyname[] =
     "Software\\Microsoft\\windows\\currentversion\\uninstall\\TeXLivexxxx";

  if (!release[0]) return FALSE;
  strncpy( keyname + strlen( keyname ) - 4, release, 4 );

  result = RegOpenKeyExA(
    HKEY_LOCAL_MACHINE,
    keyname,
    0,
    KEY_QUERY_VALUE,
    &uninst_key);
  // An additional check: compare the path of the uninstaller
  // with the path of this program, using
  // the GetFileInformationByHandle API call, this time using wchar_t
  return ( result == ERROR_SUCCESS );
}

BOOL is_elevated() {

  BOOL ret = FALSE;

  // http://stackoverflow.com/questions/8046097/
  HANDLE ProcessHandle, TokenHandle = NULL;
  TOKEN_ELEVATION Elevation;
  DWORD cbsize = sizeof( TOKEN_ELEVATION );
  ProcessHandle = GetCurrentProcess(); // no need to close afterwards
  if( OpenProcessToken( ProcessHandle, TOKEN_QUERY, &TokenHandle )) {
    if( GetTokenInformation(
        TokenHandle,
        TokenElevation, // NOT a variable!
        &Elevation,
        sizeof( Elevation ),
        &cbsize )) {
      ret = Elevation.TokenIsElevated;
      if( TokenHandle ) CloseHandle( TokenHandle );
    }
  }
  return ret;
}

// starting the real tlshell
void start_real_tlshell() {
  wchar_t * cmdln;
  STARTUPINFO si;
  PROCESS_INFORMATION pi;
  // command-line
  tl_wasprintf( &cmdln, L"\"%ls\" \"%ls\"", tclkit, tlshell );
  // create child process the normal way
  ZeroMemory( &si, sizeof(si) );
  si.cb = sizeof(si);
  si.dwFlags = STARTF_USESHOWWINDOW;
  si.wShowWindow = SW_HIDE ;
  ZeroMemory( &pi, sizeof(pi) );
  if( !CreateProcess(
    NULL,   // module name (uses command line if NULL)
    cmdln,  // command line
    NULL,   // process security atrributes
    NULL,   // thread security atrributes
    TRUE,   // handle inheritance
    0,      // creation flags, e.g. CREATE_NEW_CONSOLE, CREATE_NO_WINDOW, DETACH
    NULL,   // pointer to environment block (uses parent if NULL)
    NULL,   // starting directory (uses parent if NULL)
    &si,    // STARTUPINFO structure
    &pi )   // PROCESS_INFORMATION structure
  ) die( L"Failed to start tlshell", 1 );
  else PostQuitMessage( 0 );
}

// starting the real tlshell with an elevation prompt
void start_real_tlshell_elevated() {
  wchar_t * qtlshell;
  SHELLEXECUTEINFO shExecInfo;
  tl_wasprintf( &qtlshell, L"\"%ls\"", tlshell );
  // this is the recommended way to get an elevation prompt(!)
  shExecInfo.cbSize = sizeof(SHELLEXECUTEINFO);
  shExecInfo.fMask = SEE_MASK_NO_CONSOLE;
  shExecInfo.hwnd = NULL;
  shExecInfo.lpVerb = L"runas";
  shExecInfo.lpFile = tclkit;
  shExecInfo.lpParameters = qtlshell;
  shExecInfo.lpDirectory = NULL;
  shExecInfo.nShow = SW_HIDE;
  shExecInfo.hInstApp = NULL;
  if ( !ShellExecuteEx( &shExecInfo ))
    die( L"Failed to start tlshell elevated", 1 );
  else PostQuitMessage( 0 );
}

int WINAPI wWinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance,
    PWSTR cmdline, int iShow ) {
  DWORD l;
  wchar_t * wptr;
  wchar_t * wdummy;
  int i;

  // calculate installation root: grandparent of own directory
  // msdn is conflicted as to whether _wpgmptr is absolute;
  // don't use _get_wpgmptr
  wdummy = calloc_fatal( 32768, sizeof( wchar_t ));
  l = GetModuleFileNameW( NULL, wdummy, 32768 ); // length l excludes \0
  if ( l == 0 || GetLastError() != ERROR_SUCCESS )
    die( L"Tlshell: cannot get own path", 1 );
  for (i=0; i<3; i++) {
    wptr = wcsrchr( wdummy, L'\\' );
    if ( !wptr ) die( L"Did not find TeX Live installation;\n"
      L"this is not the right tlshell.", 0 );
    *wptr = L'\0';
  }
  // just copying, but let tl_wasprintf handle the allocation nuisance:
  tl_wasprintf( &instroot, L"%ls", wdummy );
  FREE0( wdummy );
  // locate tclkit
  tl_wasprintf( &tclkit, L"%ls\\tlpkg\\tltcl\\tclkit.exe", instroot );
  if( ! FILE_EXISTS( tclkit )) die( L"Cannot find tclkit.exe", 0 );
  // and tlshell.tcl
  tl_wasprintf(
    &tlshell, L"%ls\\texmf-dist\\scripts\\tlshell\\tlshell.tcl", instroot );
  if( ! FILE_EXISTS( tlshell )) {
    FREE0( tlshell );
    tl_wasprintf(
      &tlshell, L"%ls\\texmf-dist\\scripts\\texlive\\tlshell.tcl", instroot );
  }
  if( ! FILE_EXISTS( tlshell )) die( L"Tlshell not available", 0 );

  // read TL release from release file
  get_tl_release();
  if( ! release[0] ) {
    MessageBox( NULL, L"Problem with installation: cannot determine release!",
        L"Error", MB_OK | MB_ICONSTOP );
    return 1;
  }
  // tl_wasprintf( &wdummy, L"%ls\n%ls\n%ls", instroot, tclkit, tlshell );
  // MessageBox( NULL, wdummy, L"Info", MB_OK );
  // FREE0( wdummy );
  // MessageBoxA( NULL, release, release, 0 );
  // check for an uninstall entry under HKLM
  if ( ! in_HKLM() )
    start_real_tlshell();
  else if ( is_elevated() )
    start_real_tlshell();
  else
    start_real_tlshell_elevated();
  return 0;
}
